from typing import overload

from bento_meta.objects import Edge, Node, Property, Tag, Term

@overload
def spec_to_entity(
    hdl: str | None,
    spec: dict,
    init: dict,
    ent_cls: type[Node],
) -> Node: ...
@overload
def spec_to_entity(
    hdl: str | None,
    spec: dict,
    init: dict,
    ent_cls: type[Edge],
) -> Edge: ...
@overload
def spec_to_entity(
    hdl: str | None,
    spec: dict,
    init: dict,
    ent_cls: type[Property],
) -> Property: ...
@overload
def spec_to_entity(
    hdl: str | None,
    spec: dict,
    init: dict,
    ent_cls: type[Term],
) -> Term: ...
@overload
def spec_to_entity(
    hdl: str | None,
    spec: dict,
    init: dict,
    ent_cls: type[Tag],
) -> Tag: ...



"""
DiffSummary class summarizes differences between two MDFs in plain text.

Typical usage example:
    summary = DiffSummary(diff).create_summary()
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from bento_meta.objects import Term

if TYPE_CHECKING:
    from bento_mdf.diff import Diff


class DiffSummary:
    """
    This class provides methods to summarize the differences between two MDFs.

    The create_summary method generates the summary string.
    """

    def __init__(self, diff: Diff) -> None:
        """Initialize the DiffSummary object from a Diff."""
        self.diff = diff

    def create_overall_summary(self) -> str:
        """Create a broad overall summary of the differences between two MDFs."""
        summary_added_removed = []
        summary_changed = []
        for ent_type, diffs in self.diff.result.items():
            ents_with_attr_changes = 0
            attr_changes = 0
            for action, diff in diffs.items():
                if action in ["added", "removed"]:
                    count = self.count_items(ent_type, action)
                    if count > 0:
                        summary_added_removed.append(
                            f"{count} {ent_type[:-1]}(s) {action}",
                        )
                else:  # action == "changed"
                    ents_with_attr_changes += len(diffs[action])
                    attr_changes += sum(len(d) for d in diff.values())
            if ents_with_attr_changes != 0 and attr_changes != 0:
                summary_changed.append(
                    f"{attr_changes} attribute(s) changed for "
                    f"{ents_with_attr_changes} {ent_type[:-1]}(s)",
                )

        return "; ".join(summary_added_removed + summary_changed)

    def create_detailed_summary(self) -> str:
        """Create a detailed summary of the differences between two MDFs."""
        detailed_summary = []
        for ent_type, diffs in self.diff.result.items():
            for action in diffs:
                items = self.get_items(ent_type, action)
                if not items:
                    continue
                if action in ["added", "removed"]:
                    for item in items:
                        detail = self.format_add_rem_detail(ent_type, action, item)
                        detailed_summary.append(detail)
                else:  # action == "changed"
                    for item in items:
                        if ent_type != "terms":
                            continue
                        detail = self.format_changed_detail(ent_type, item)
                        detailed_summary.append(detail)
        return "\n".join(detailed_summary)

    def count_items(self, ent_type: str, action: str) -> int:
        """Count the number of items in the diff dictionary."""
        items = self.get_items(ent_type, action)
        if not items:
            return 0
        return len(items)

    def get_items(self, ent_type: str, action: str) -> list | None:
        """Get the items from the diff dictionary."""
        items = self.diff.result.get(ent_type, {}).get(action, {})
        if isinstance(items, dict):
            return list(items.items())
        return items

    def get_detail_by_ent_type(
        self,
        ent_type: str,
        entk: str | tuple[str, str] | tuple[str, str, str],
    ) -> str:
        """Return formatted string for detailed summary based on entity type and key."""
        props_key_len = 2
        edge_key_len = 3
        ent_type_details = {
            "nodes": f"'{entk}'",
            "edges": f"'{entk[0]}' with src: '{entk[1]}' and dst: '{entk[2]}'"
            if len(entk) == edge_key_len
            else "Invalid key format for edges",
            "props": f"'{entk[1]}' with parent: '{entk[0]}'"
            if len(entk) == props_key_len
            else "Invalid key format for 'props'",
            "terms": f"'{entk[0]}' with origin: '{entk[1]}'",
        }
        return ent_type_details[ent_type]

    def format_entity_annotation(self, action: str, item: tuple) -> str:
        """Check if term annotates an entity and format accordingly."""
        change_detail = ""
        if action in ["added", "removed"] and not isinstance(item[1], Term):
            if not item[1].concept or not item[1].concept.terms:
                return ""
            term_annos = item[1].concept.terms
            term_anno_str = ", ".join(
                [
                    f"'{term_annos[_].value}' with origin '{term_annos[_].origin_name}'"
                    for _ in term_annos
                ],
            )
            return f". {item[1].get_label().capitalize()} annotated by: {term_anno_str}"
        if action == "changed":
            action = "added"
            for attr, changes in item[1].items():
                change_detail += (
                    f". Attribute: '{attr}' updated from "
                    f"'{changes['removed']}' to '{changes['added']}'"
                )
        for anno_ent_type, anno_changes in self.diff.annotations.items():
            for entk, anno_change in anno_changes.items():
                anno_item = anno_change.get(action, {})
                if not anno_item:
                    continue
                for termk in anno_item:
                    if termk != item[0]:
                        continue
                    return (
                        f" which annotates {anno_ent_type[:-1]}: "
                        f"{self.get_detail_by_ent_type(anno_ent_type, entk)}"
                        f"{change_detail}"
                    )
        return ""

    def format_add_rem_detail(
        self,
        ent_type: str,
        action: str,
        item: tuple,
    ) -> str:
        """Format detailed summary for added/removed entities depending on type."""
        detail = f"- {action.capitalize()} {ent_type[:-1]}: "
        if ent_type == "nodes":
            detail += f"'{item[0]}'"
        elif ent_type == "edges":
            detail += f"'{item[0][0]}' with src: '{item[0][1]}' and dst: '{item[0][2]}'"
        elif ent_type == "props":
            detail += f"'{item[0][1]}' with parent: '{item[0][0]}'"
        elif ent_type == "terms":
            detail += f"'{item[0][0]}' with origin: '{item[0][1]}'"
        detail += self.format_entity_annotation(action, item)
        return detail

    def format_changed_detail(self, ent_type: str, item: tuple) -> str:
        """Format detailed summary for changed entities depending on type."""
        if ent_type != "terms":
            return ""
        detail = f"- Changed {ent_type[:-1]}: "
        detail += f"'{item[0][0]}' with origin: '{item[0][1]}'"
        detail += self.format_entity_annotation("changed", item)
        return detail

    def create_summary(self) -> str | None:
        """Generate a string summarizing the differences between two MDFs."""
        if not self.diff.result:
            return None

        overall_summary = self.create_overall_summary()
        detailed_summary = self.create_detailed_summary()

        return f"{overall_summary}\n{detailed_summary}"



from __future__ import annotations
import re
import sys
import importlib.util
from functools import cache
from pathlib import Path
from .reader import MDFReader
from bento_meta.objects import Property
from tempfile import NamedTemporaryFile
from jinja2 import Environment, PackageLoader
from typing import Any, List, NoReturn #, TYPE_CHECKING
from datetime import datetime
from pydantic import BaseModel, TypeAdapter, ValidationError, AnyUrl
from pydantic.json_schema import GenerateJsonSchema
from pdb import set_trace

jenv = Environment(
    loader=PackageLoader("bento_mdf", package_path="mdf/templates"),
    trim_blocks=True,
)

# jinja helpers    

def toCamelCase(val : str) -> str:
    return "".join([x.capitalize() for x in val.split("_")])


def to_snakecase(val : str) -> str:
    return re.sub("\\W", "_", val).lower()


def to_unit_types(unitstr : str, typ : type(int) | type(float)) -> List[str]:
    return [f"Annotated[{typ.__name__}, Unit('{u}')]" for u in unitstr.split(';')]


def maybe_optional(val : str, prop : Property):
    if prop.is_required:
        return val
    else:
        return f"Optional[{val}]"

def maybe_list(val : str, prop : PropertyO):
    if prop.value_domain == "list":
        return f"List[{val}]"
    else:
        return val


jenv.filters['toCamelCase'] = toCamelCase
jenv.filters['to_snakecase'] = to_snakecase
jenv.filters['to_unit_types'] = to_unit_types
jenv.filters['maybe_optional'] = maybe_optional
jenv.filters['maybe_list'] = maybe_list

class GenerateQualJsonSchema(GenerateJsonSchema):
    # override to add $schema tag
    def generate(self, schema, mode='validation'):
        json_schema = super().generate(schema, mode=mode)
        json_schema['$schema'] = self.schema_dialect
        return self.sort(json_schema)


class MDFDataValidator:
    typemap = {
        "boolean": bool,
        "datetime": datetime,
        "integer": int,
        "number": float,
        "string": str,
        "regexp": str,
        "url": AnyUrl,
        "TBD": Any,
        }

    def __init__(self, mdf: MDFReader):
        self.model = mdf.model
        self._pymodel = None
        self._module = None
        self._node_classes = []
        self._enum_classes = []
        self._validation_errors = None
        self.generate_data_model()
        self.import_data_model()
        
    @property
    def data_model(self) -> str:
        return self._pymodel

    @property
    def module(self):
        return self._module

    @property
    def model_class(self) -> str:
        return self.module and self.module.__name__

    @property
    def node_classes(self) -> List:
        return self._node_classes

    @property
    def enum_classes(self) -> List:
        return self._enum_classes

    @property
    def last_validation_errors(self) -> dict | None:
        return self._validation_errors
    
    def generate_data_model(self) -> NoReturn:
        """
        Generates Pydantic classes for each node in the MDF model.
        """
        # write down classes to be generated
        for node in self.model.nodes.values():
            self._node_classes.append(toCamelCase(node.handle))
            for pr in node.props.values():
                if pr.value_domain == 'value_set' or pr.item_domain == 'value_set':
                    if pr.value_set.url:
                        self._enum_classes.append("{}EnumURL".format(toCamelCase(pr.handle)))
                    elif pr.value_set.path:
                        self._enum_classes.append("{}EnumPath".format(toCamelCase(pr.handle)))
                    else:
                        self._enum_classes.append("{}Enum".format(toCamelCase(pr.handle)))
        self._node_classes.sort()
        self._enum_classes.sort()
        template = jenv.get_template("pymodel.py.jinja2")
        self._pymodel = template.render(model=self.model, typemap=self.typemap)

    def import_data_model(self) -> NoReturn:
        """
        Imports model classes as a module '<model handle>Data'
        :returns: module object
        """
        if not self.data_model:
            return
        with NamedTemporaryFile(mode="w+", suffix=".py", delete=False) as modf:
            modname = "{}Data".format(self.model.handle)
            print(self.data_model, file=modf)
            modf.close()
            spec = importlib.util.spec_from_file_location(modname, modf.name)
            module = importlib.util.module_from_spec(spec)
            sys.modules[modname] = module
            spec.loader.exec_module(module)
            self._module = module
            Path(modf.name).unlink()

    @cache
    def model_of(self, clsname : str):
        if clsname != self.model_class and clsname not in self.node_classes and clsname not in self.enum_classes:
            raise RuntimeError(f"Validation model does not contain class '{clsname}'")
        return eval("self.module.{}".format(clsname))

    @cache
    def fields_of(self, clsname : str) -> List[str]:
        if clsname != self.model_class and clsname not in self.node_classes:
            raise RuntimeError(f"Validation model does not contain node class '{clsname}'")
        return [x for x in self.model_of(clsname).model_fields]

    def props_of(self, clsname : str) -> List[str]:
        return self.fields_of(clsname)
    
    @cache
    def validator(self, clsname : str):
        """
        Return a validator function appropriate to the class named 'clsname'
        """
        model = self.model_of(clsname)
        if issubclass(model, BaseModel):
            return model.model_validate
        else:
            return TypeAdapter(model).validate_python
        
    def json_schema(self, clsname : str) -> dict | list:
        """
        Return a jsonable object representing a JSONSchema that can validate
        the given model class.
        """

        model = self.model_of(clsname)
        if issubclass(model, BaseModel):
            return model.model_json_schema(
                schema_generator=GenerateQualJsonSchema)
        else:
            return TypeAdapter(model).json_schema(
                schema_generator=GenerateQualJsonSchema)
        
    def validate(self, clsname : str, data : dict | List[dict], strict : bool = False,
                 verbose : bool = False) -> bool:
        """
        Validate a dict or list of dicts against a given model class.
        Returns true if all items are valid, false otherwise.
        If items fail validation, The attribute 'last_validation_errors' will contain
        a dict whose keys are the index of the item in the submitted data list, and values
        which are lists of specific errors found in the item.
        ('last_validation_errors' is reset to None if validation succeeds.)
        """
        dta = []
        result = True
        self._validation_errors = {}
        if isinstance(data, dict):
            dta = [data]
        else:
            dta = data
        valf = self.validator(clsname)
        for i, rec in enumerate(dta):
            try:
                valf(rec, strict=strict)
            except ValidationError as e:
                result = False
                if verbose:
                    print(e.title, file=sys.stderr)
                self._validation_errors[i] = e.errors()
        if result:
            self._validation_errors = None
        return result




#!/usr/bin/env python
import sys

sys.path.insert(0, "..")
import argparse
import getpass
from sys import stderr

from bento_mdf.mdf import MDF
from bento_meta.mdb import (
    WriteableMDB,
    load_mdf,
    load_model_statements,
    make_nanoid,
)

parser = argparse.ArgumentParser(description="Load model in MDF into an MDB")
parser.add_argument("files", nargs="*", metavar="MDF-FILE", help="MDF file(s)/url(s)")
parser.add_argument(
    "--commit",
    default="",
    help="commit SHA1 for MDF instance (if any)",
    required=True,
)
parser.add_argument("--handle", help="model handle")
parser.add_argument("--user", help="MDB username")
parser.add_argument("--passw", help="MDB password")
parser.add_argument(
    "--bolt",
    metavar="BoltURL",
    help="MDB Bolt url endpoint (specify as 'bolt://...')",
)
parser.add_argument("--put", action="store_true", help="Load model to database")
parser.add_argument(
    "--make-nanoids",
    action="store_true",
    help="Add new nanoids to graph nodes",
)
# example:
# args = parser.parse_args([
#     "https://raw.githubusercontent.com/CBIIT/icdc-model-tool/master/model-desc/icdc-model.yml",
#     "https://raw.githubusercontent.com/CBIIT/icdc-model-tool/master/model-desc/icdc-model-props.yml",
#     "--commit",
#     "a4aa9a43b9ad2087638ceaeef50d4a22a4b9b959",
#     "--handle",
#     "ICDC",
#     "--bolt",
#     "bolt://localhost:7687",
#     "--user",
#     "neo4j",
#     "--passw",
#     getpass.getpass()
#     ])

args = parser.parse_args()

if not args.files:
    parser.print_help(file=stderr)
    parser.exit(1)

if args.put and not args.bolt:
    print("error: --bolt and --user args required to put to database", file=stderr)
    parser.exit(2)

if args.put and not args.passw:
    args.passw = getpass.getpass()

print("load model from MDFs", file=stderr)
mdf = MDF(*args.files, handle=args.handle, _commit=args.commit, raise_error=True)
model = mdf.model

if args.put:
    print("Put model to DB", file=stderr)
    mdb = WriteableMDB(uri=args.bolt, user=args.user, password=args.passw)
    model.mdb = mdb
    load_mdf(mdf, model.mdb)
    if args.make_nanoids:
        print("Add nanoids to nodes", file=stderr)
        with mdb.driver.session() as s:
            result = s.run(
                "match (n {_commit:$commit}) where not exists(n.nanoid) "
                "with n limit 1 set n.nanoid=$nanoid return n",
                {"commit": args.commit, "nanoid": make_nanoid()},
            )
            while result.peek():
                result = s.run(
                    "match (n {_commit:$commit}) where not exists(n.nanoid) "
                    "with n limit 1 set n.nanoid=$nanoid return n",
                    {"commit": args.commit, "nanoid": make_nanoid()},
                )
else:
    for s in load_model_statements(model, args.commit):
        print(s)



"""
Load Model Description Format (MDF) files to bento-meta Models.

This module contains :class:`MDF`, a class for reading a graph data model in
Model Description Format into a :class:`bento_meta.model.Model` object.
"""

from __future__ import annotations

import logging
import re
from collections import ChainMap
from pathlib import Path
from tempfile import TemporaryFile
from urllib.parse import urlparse

import requests
from bento_meta.entity import ArgError, Entity
from bento_meta.model import Model
from bento_meta.objects import Edge, Node, Property, Term
from nanoid import generate
from tqdm import tqdm

from bento_mdf.mdf.convert import spec_to_entity
from bento_mdf.validator import MDFValidator


def make_nano() -> str:
    """Generate a 6-character alphanumeric string."""
    return generate(
        size=6,
        alphabet="abcdefghijkmnopqrstuvwxyzABCDEFGHJKMNPQRSTUVWXYZ0123456789",
    )


class MDFReader:
    """MDF class for reading MDF files into a bento-meta Model."""

    def __init__(
        self,
        *yaml_files: str | Path | list[str | Path],
        handle: str | None = None,
        model: Model | None = None,
        _commit: str | None = None,
        mdf_schema: str | Path | None = None,
        raise_error: bool = False,
        logger: logging.Logger | None = None,
    ) -> None:
        """
        Create a :class:`Model` from MDF YAML files/Write a :class:`Model` to YAML.

        :param str|file|url *yaml_files: MDF filenames or file objects,
        in desired merge order
        :param str handle: Handle (name) for the resulting Model
        :param :class:`Model` model: Model to convert to MDF
        :param boolean raise_error: raise on error if True
        :param :class:`logging.Logger` logger: Python logger (suitable default)
        :attribute model: the :class:`bento_meta.model.Model` created
        """
        if model and not isinstance(model, Model):
            msg = "arg model= must be a Model instance"
            raise ArgError(msg)

        self.files = yaml_files
        self.mdf = {}
        self.mdf_schema = mdf_schema
        self._model = model
        self._commit = _commit
        self._annotations = {}
        self._terms = {}
        self._props = {}
        self.version = None
        self.uri = None
        self.logger = logger or logging.getLogger(__name__)
        self.create_model_success = False
        if model:
            self.handle = model.handle
        else:
            self.handle = handle
        if self.files:
            self.load_yaml()
            self.create_model(raise_error=raise_error)
        elif not model:
            self.logger.warning("No MDF files or model provided to constructor")

    @property
    def model(self) -> Model:
        """The :class:`bento_meta.model.Model` object created from the MDF input."""
        if not self._model:
            msg = "Can't fetch model from MDF."
            self.logger.error(msg)
            raise ArgError(msg)
        return self._model

    def load_yaml(self, *, verify: bool = True) -> None:
        """Validate and load YAML files or open file handles specified in constructor."""
        vargs = []
        for f in self.files:
            if isinstance(f, str) and re.match("(?:file|https?)://", f):
                self.load_yaml_vargs_from_url(vargs, f, verify=verify, raise_error=True)
            elif isinstance(f, str) and Path(f).exists():
                fh = Path(f).open(encoding="utf-8")
                vargs.append(fh)
            else:  # assume file-like object
                vargs.append(f)

        v = MDFValidator(self.mdf_schema, *vargs, raise_error=True)
        self.mdf_schema = v.load_and_validate_schema()
        self.mdf = v.load_and_validate_yaml()
        if not self.mdf_schema:
            msg = "Error loading & validating MDF schema"
            raise ValueError(msg)
        if not self.mdf:
            msg = "Error loading & validating YAML instance"
            raise ValueError(msg)
        self.mdf = self.mdf.as_dict()

        for fh in vargs:
            if hasattr(fh, "close") and callable(fh.close):
                fh.close()

    def load_yaml_vargs_from_url(
        self,
        vargs: list,
        url: str,
        *,
        verify: bool = True,
        raise_error: bool = False,
    ) -> None:
        """Load YAML from a URL. Converts GitHub repo URLs to raw URLs."""
        raw_url = convert_github_url(url)
        try:
            response = requests.get(raw_url, verify=verify, timeout=10)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            msg = f"Fetching url {raw_url} raised exception {e}"
            self.logger.error(msg)
            if raise_error:
                raise ArgError(msg) from e
            return
        response.encoding = "utf8"
        fh = TemporaryFile()
        for chunk in response.iter_content(chunk_size=128):
            fh.write(chunk)
        fh.seek(0)
        vargs.append(fh)

    def create_model(self, *, raise_error: bool = False) -> Model:
        """
        Create :class:`Model` instance from loaded YAML.

        Handling enumerated acceptable values ("value sets"):

        If the input MDF does not have a Terms: section, the strings listed under
        an Enum: key are interpreted as literal Term values (i.e., the expected data),
        and the Origin of each Term is set as the model handle (i.e.,the model is itself
        the authority defining the acceptable values)

        However, if the input MDF contains a Terms: section, then each string of an Enum:
        array is first treated as a Term handle, and is used to look up a corresponding Term
        in the the Terms: section. If found, the definition in the Terms: section is used.
        If no referenced Term is found, then the string is treated
        as a literal Term value with the model as origin, as in the previous paragraph.

        :param boolean raiseError: Raise if MDF errors found
        Note: This is brittle, since the syntax of MDF is hard-coded into this method.
        """
        self.create_model_success = True
        if not self.mdf.keys():
            msg = "attribute 'mdf' not set - are yamls loaded?"
            raise ValueError(msg)
        self.version = self.mdf.get("Version")
        self.uri = self.mdf.get("URI")

        if self.handle:
            self._model = Model(handle=self.handle, version=self.version, uri=self.uri)
        elif self.mdf.get("Handle"):
            self.handle = self.mdf["Handle"]
            self._model = Model(
                handle=self.mdf["Handle"],
                version=self.version,
                uri=self.uri,
            )
        else:
            self.logger.error("Model handle not present in MDF nor provided in args")
            self.create_model_success = False

        self.create_terms()  # create terms first, if any -- properties depend on these
        self.create_nodes()
        self.create_edges()
        self.create_props()
        if raise_error and not self.create_model_success:
            msg = "MDF errors found; see log output."
            raise RuntimeError(msg)

        # now add the terms collected in self._terms to the model (self._model)...

        return self.model

    def create_terms(self) -> None:
        """Create terms from loaded YAML."""
        if "Terms" not in self.mdf:
            return
        for t_hdl, spec in tqdm(self.mdf["Terms"].items()):
            if "Value" not in spec:
                self.logger.error(
                    "Term specs must have a Value key and a non-null string value"
                    "(term '%s')",
                    t_hdl,
                )
                self.model_create_success = False
            if "Origin" not in spec:
                self.logger.warning(
                    f"No Origin provided for term '{t_hdl}'",
                )
            term = spec_to_entity(t_hdl, spec, {"_commit": self._commit}, Term)
            self._terms[term.handle] = term

    def create_nodes(self) -> None:
        """Create nodes from loaded YAML."""
        for n, spec in self.mdf["Nodes"].items():
            node = self.model.add_node(
                spec_to_entity(
                    n,
                    spec,
                    {"model": self.handle, "_commit": self._commit},
                    Node,
                ),
            )
            if "Term" in spec:
                self.annotate_entity_from_mdf(node, spec["Term"])

    def create_edges(self) -> None:
        """Create edges from loaded YAML."""
        for e, spec in self.mdf["Relationships"].items():
            for ends in spec["Ends"]:
                for end in [ends["Src"], ends["Dst"]]:
                    if end not in self.model.nodes:
                        self.logger.warning(
                            "No node '%s' defined for edge spec '%s' from '%s' to '%s'",
                            end,
                            e,
                            ends["Src"],
                            ends["Dst"],
                        )
                        continue
                if not spec.get("Mul"):
                    self.logger.warning(
                        "edge '%s' from '%s' to '%s' "
                        "does not specify a multiplicity",
                        e,
                        ends["Src"],
                        ends["Dst"],
                    )
                if spec.get("Mul") not in (
                    "one_to_one",
                    "one_to_many",
                    "many_to_one",
                    "many_to_many",
                ):
                    self.logger.warning(
                        "edge '%s' from '%s' to '%s'"
                        " has non-standard multiplicity '%s'",
                        e,
                        ends["Src"],
                        ends["Dst"],
                        spec.get("Mul"),
                    )
                # if tags are set in the Ends entry, these will be attached
                # to the Edge, _rather than_ tags set in the "edge key" or spec
                # level
                spec["Tags"] = ends.get("Tags") or spec.get("Tags")
                edge = self.model.add_edge(
                    spec_to_entity(
                        e,
                        spec,
                        {
                            "model": self.handle,
                            "_commit": self._commit,
                            "src": self.model.nodes[ends["Src"]],
                            "dst": self.model.nodes[ends["Dst"]],
                            "multiplicity": ends.get("Mul")
                            or spec.get("Mul")
                            or Edge.default("multiplicity"),
                        },
                        Edge,
                    ),
                )
                term = ends.get("Term") or spec.get("Term")
                if term:
                    self.annotate_entity_from_mdf(edge, term)

    def create_props(self) -> None:
        """Create properties from loaded YAML."""
        yunps = self.mdf.get("UniversalNodeProperties")
        yurps = self.mdf.get("UniversalRelationshipProperties")

        propnames = {}
        for ent in ChainMap(self.model.nodes, self.model.edges).values():
            if isinstance(ent, Node):
                pnames = self.mdf["Nodes"][ent.handle]["Props"] or []
                if yunps:  # universal node props
                    pnames.extend(yunps["mayHave"] if yunps.get("mayHave") else [])
                    pnames.extend(yunps["mustHave"] if yunps.get("mustHave") else [])
                if pnames:
                    propnames[ent] = pnames
            elif isinstance(ent, Edge):
                # props elts appearing in Ends hash take precedence over
                # Props elt in the handle's hash
                (hdl, src, dst) = ent.triplet
                ends = [
                    e
                    for e in self.mdf["Relationships"][hdl]["Ends"]
                    if e["Src"] == src and e["Dst"] == dst
                ]
                if len(ends) > 1:
                    self.logger.warning(
                        "edge '%s' has more than one Ends pair Src:'%s',Dst:'%s'",
                        hdl,
                        src,
                        dst,
                    )
                end = ends[0]

                # note the end-specified props _replace_ the edge-specified props,
                # they are not merged:
                pnames = (
                    end.get("Props")
                    or self.mdf["Relationships"][hdl].get("Props")
                    or []
                )
                if yurps:  # universal relationship props
                    pnames.extend(yurps["mayHave"] if yurps.get("mayHave") else [])
                    pnames.extend(yurps["mustHave"] if yurps.get("mustHave") else [])
                if pnames:
                    propnames[ent] = pnames
            else:
                self.logger.error(
                    "Unhandled entity type %s for properties",
                    type(ent).__name__,
                )
                self.create_model_success = False
        prop_of = {}
        for ent in propnames:
            for p in propnames[ent]:
                if prop_of.get(p):
                    prop_of[p].append(ent)
                else:
                    prop_of[p] = [ent]
        propdefs = self.mdf["PropDefinitions"]
        defns_for = set(propdefs.keys())
        for pname in prop_of:
            for ent in prop_of[pname]:
                force = False
                # see if a qualified name is defined in propdefs:
                key = ent.handle + "." + pname
                spec = propdefs.get(key)
                if spec:
                    # force creation of new prop for a explicitly qualified MDF property
                    force = True
                else:
                    key = pname
                    spec = propdefs.get(pname)
                if not spec:
                    self.logger.warning(
                        "property '%s' does not have a corresponding "
                        "propdef for entity '%s'",
                        pname,
                        ent.handle,
                    )
                    continue
                if key in defns_for:
                    defns_for.remove(key)
                prop = self.create_or_merge_prop_from_mdf(
                    spec,
                    p_hdl=pname,
                    force_create=force,
                )
                self.model.add_prop(ent, prop)
                ent.props[prop.handle] = prop
        if defns_for:
            self.logger.warning(
                "No properties in model corresponding to the following "
                "PropDefintions: %s",
                defns_for,
            )

    def create_or_merge_prop_from_mdf(
        self,
        spec: dict,
        p_hdl: str,
        *,
        force_create: bool,
    ) -> Property:
        if not force_create and (self.handle, p_hdl) in self._props:
            pass
        else:
            prop = spec_to_entity(
                p_hdl,
                spec,
                {"handle": p_hdl, "model": self.handle, "_commit": self._commit},
                Property,
            )
            if prop.value_set and (prop.value_set.path is not None
                                   or prop.value_set.url is not None):  # enum as reference
                self.merge_enum_reference(prop)
            if prop.value_set and prop.value_set._commit == "dummy":
                terms = []
                # merge terms references in enums into terms defined
                # in a separate Terms: section
                for t in prop.value_set.terms:
                    if self._terms.get(t):
                        terms.append(self._terms[t])
                    else:
                        terms.append(prop.value_set.terms[t])
                # allow bento_meta.model machinery to create
                # the actual value_set object and store
                # terms in Model object:
                self.add_terms_to_model_prop(prop, terms)

            if "Term" in spec:
                self.annotate_entity_from_mdf(prop, spec["Term"])
            if force_create:
                return prop
            self._props[(prop.model, prop.handle)] = prop
        return self._props[(self.handle, p_hdl)]

    def load_enum_reference(
        self,
        enum_ref: str,
    ) -> dict[str, dict[str, list[str] | dict[str, str]]]:
        """Load enum from a reference (path or url, yaml file or list of strings)."""
        if re.match("^/", enum_ref):  # looks like a path
            enum_path = (Path.cwd() / Path(enum_ref.lstrip("/"))).resolve()
            if not enum_path.exists():
                self.logger.error("Enum reference path '%s' does not exist", enum_path)
                return {}
            with Path(enum_path).open() as f:
                v = MDFValidator(None, f)
                enum_mdf = v.load_and_validate_yaml()
                if not enum_mdf:
                    self.logger.error("Error loading enum from path '%s'", enum_path)
                    return {}
                return enum_mdf.as_dict()  # type: ignore reportReturnType
        if re.match("(?:file|https?)://", enum_ref):  # looks like a url
            vargs = []
            self.load_yaml_vargs_from_url(vargs, enum_ref, raise_error=False)
            v = MDFValidator(None, *vargs)
            enum_mdf = v.load_and_validate_yaml()
            if not enum_mdf:
                self.logger.error("Error loading enum from url '%s'", enum_ref)
                return {}
            return enum_mdf.as_dict()  # type: ignore reportReturnType
        return {}

    def merge_enum_reference(self, prop: Property) -> None:
        """
        Merge enum from a reference (path or url to yaml file or list of strings).

        Adds terms to the property value set.
        """
        if not prop.value_set or not (prop.value_set.path or prop.value_set.url):
            self.logger.error("No enum reference in property '%s'", prop.handle)
            return
        enum = self.load_enum_reference(prop.value_set.path or prop.value_set.url)
        enum_values = enum.get("PropDefinitions", {}).get(prop.handle, [])
        enum_terms = enum.get("Terms", {})
        if not enum_values:
            self.logger.error(
                "No enum at reference '%s'",
                prop.value_set.path or prop.value_set.url,
            )
            return
        specs = {val: {"Value": val} for val in enum_values}
        if enum_terms:  # merge term definitions with enum values
            specs.update(
                {val: enum_terms.get(val, {"Value": val}) for val in enum_values},
            )
        for spec in specs.values():
            if "Origin" in spec:
                continue
            spec["Origin"] = prop.model
        terms = [
            spec_to_entity(None, spec, {"_commit": "dummy"}, Term)
            for spec in specs.values()
        ]
        self.add_terms_to_model_prop(prop, terms)

    def add_terms_to_model_prop(self, prop: Property, terms: list[Term]) -> None:
        """Add terms to a model property & handles list type props with value sets."""
        if prop.value_domain == "list" and prop.item_domain == "value_set":
            # kludge so Model.add_terms works with list type props with val sets
            prop.value_domain = "value_set"
            self.model.add_terms(prop, *terms)
            prop.value_domain = "list"
        else:
            self.model.add_terms(prop, *terms)

    def annotate_entity_from_mdf(self, ent: Entity, yterm_list: list) -> None:
        """Annotate an entity from a list of term references in MDF."""
        for spec in yterm_list:
            if "Value" not in spec:
                self.logger.error(
                    "Term specs must have a Value key and a non-null string value, "
                    "in entity '%s'",
                    ent.handle,
                )
                return
            if "Origin" not in spec:
                self.logger.warning(
                    "No Origin provided for Term annotation in entity '%s'",
                    ent.handle,
                )
                spec["Origin"] = self.handle
            term = spec_to_entity(None, spec, {"_commit": self._commit}, Term)
            # merge or record term
            if not self._annotations.get((term.handle, term.origin_name)):
                self._annotations[(term.handle, term.origin_name)] = term
            else:
                term = self._annotations[(term.handle, term.origin_name)]

            self.model.annotate(ent, term)
            if self._commit and not ent.concept._commit:
                ent.concept._commit = self._commit


def convert_github_url(url: str) -> str:
    """Convert a GitHub blob URL to a raw URL."""
    parsed_url = urlparse(url)
    parts = parsed_url.path.strip("/").split("/")
    if parsed_url.netloc != "github.com" or len(parts) < 4 or parts[2] != "blob":
        return url  # not a GitHub blob URL
    user, repo, _, branch = parts[:4]
    file_path = "/".join(parts[4:])
    return f"https://raw.githubusercontent.com/{user}/{repo}/{branch}/{file_path}"



import logging
import yaml
from collections import Counter
from functools import reduce
from bento_meta.entity import ArgError, Entity
from bento_meta.model import Model
from bento_meta.objects import Edge, Node, Property, Term
from .convert import entity_to_spec
from pdb import set_trace


class MDFWriter(object):
    def __init__(self, model=None):
        self.mdf = {
            "Nodes": {},
            "Relationships": {},
            "PropDefinitions": {},
            "Terms": {},
            "Handle": None,
            "Version": None,
            "URI": None,
        }
        self.logger = logging.getLogger(__name__)
        self.model = model
        if model:
            self.write_mdf()
        pass
    
    def write_mdf(self, file=None):
        """
        Use a :class:`Model` to create model description file (MDF) formatted dict
        :param :class:`Model` model: Model to convert 
        :param str|file file: File name or object to write to (default is None; just return the MDF as dict)
        :returns: MDF as dict
        """
        self.mdf["Handle"] = self.model.handle or "NEED_MODEL_HANDLE"
        self.mdf["Version"] = self.model.version or "NEED_MODEL_VERSION"
        if self.model.uri:
            self.mdf["URI"] = self.model.uri

        for nd in sorted(self.model.nodes):
            node = self.model.nodes[nd]
            self.mdf["Nodes"][nd] = entity_to_spec(node)

        for pr in sorted(self.model.props):
            prop = self.model.props[pr]
            self.mdf["PropDefinitions"][prop.handle] = entity_to_spec(prop)

        edge_specs = {}
        for rl in self.model.edges.values():
            if rl.handle in edge_specs:
                edge_specs[rl.handle].append(entity_to_spec(rl))
            else:
                edge_specs[rl.handle] = [entity_to_spec(rl)]

        for hdl in edge_specs:
            top = {}
            # determine default Mul 
            muls = Counter([x.get("Mul") for x in edge_specs[hdl]])
            dfMul = muls.most_common(1)[0][0]
            top["Mul"] = dfMul or Edge.default("multiplicity")
            # raise common Desc
            top["Desc"] = None
            descs = Counter([x.get("Desc") for x in edge_specs[hdl] if x.get("Desc")])
            if descs.total() == len(edge_specs[hdl]):
                dfDesc = descs.most_common(1)[0][0]
                top["Desc"] = dfDesc
            if not top["Desc"]:
                del top["Desc"]
            # merge props - this logic raises all Props specified in
            # Ends members to the Relationship[<handle>] level
            # (MDF schema does not currently allow Props to be specified
            #  at the End level)
            propsets = [set(x.get("Props")) for x in edge_specs[hdl]]
            if not any([len(s) > 0 for s in propsets]):
                top["Props"] = "null"
            else:
                mrgd = list(reduce(lambda x, y: x | y, propsets))
                top["Props"] = mrgd
            top["Ends"] = []
            for spec in edge_specs[hdl]:
                if spec["Mul"] == top["Mul"]:
                    del spec["Mul"]
                if spec.get("Desc") and top.get("Desc"):
                    if spec["Desc"] == top["Desc"]:
                        del spec["Desc"]
                if spec.get("Props"):
                    del spec["Props"]
                top["Ends"].append(spec)
            self.mdf["Relationships"][hdl] = top

        # collect Terms?
        for tm in sorted(self.model.terms):
            term = self.model.terms[tm]
            self.mdf["Terms"][term.handle] = entity_to_spec(term)
        
        if file:
            fh = file
            if isinstance(file, str):
                fh = open(file, "w")
            yaml.dump(self.mdf, stream=fh, indent=4)

        return self.mdf





"""YAML loader for MDF files with methods to check for duplicate keys and elements."""

from typing import Any

from yaml.constructor import ConstructorError
from yaml.loader import SafeLoader
from yaml.nodes import MappingNode, SequenceNode


class MDFLoader(SafeLoader):
    """
    Safe YAML loader for MDF files.

    Adds methods to check for duplicate keys and elements.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize the MDFLoader."""
        super().__init__(*args, **kwargs)

    def construct_mapping(
        self,
        node: MappingNode,
        *,
        deep: bool = False,
    ) -> dict:
        """Construct a mapping node with duplicate keys check."""
        if not isinstance(node, MappingNode):
            raise ConstructorError(
                None,
                None,
                "expected a mapping node, but found %s" % node.id,
                node.start_mark,
            )
        mapping = {}

        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise ConstructorError(
                    "while constructing a mapping",
                    node.start_mark,
                    "found unacceptable key (%s)" % exc,
                    key_node.start_mark,
                ) from exc
            value = self.construct_object(value_node, deep=deep)
            if key in mapping:
                raise ConstructorError(
                    "while constructing a mapping",
                    node.start_mark,
                    "found duplicated key (%s)" % key,
                    key_node.start_mark,
                )
            mapping[key] = value
        return mapping

    def construct_sequence(
        self,
        node: SequenceNode,
        *,
        deep: bool = False,
    ) -> list:
        """Construct a list of sequence nodes with duplicate elements check."""
        if not isinstance(node, SequenceNode):
            raise ConstructorError(
                None,
                None,
                "expected a sequence node, but found %s" % node.id,
                node.start_mark,
            )
        elts = set()
        for c in node.value:
            if isinstance(c.value, str):  # just check lists of strings
                if c.value in elts:
                    raise ConstructorError(
                        "while constructing a sequence",
                        node.start_mark,
                        "found duplicated element (%s)" % c.value,
                    )
                elts.add(c.value)
        return [self.construct_object(child, deep=deep) for child in node.value]



"""
bento_mdf.mdf.convert
=====================

Utilities for converting MDF YAML to bento-meta objects
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from urllib.parse import unquote

from bento_meta.objects import Edge, Node, Property, Tag, Term, ValueSet
from pdb import set_trace

if TYPE_CHECKING:
    from bento_meta.entity import Entity


def to_snake_case(string: str) -> str:
    """Convert string to snake case representation."""
    string = string.replace(" ", "_")
    string = re.sub(r"(?<=[a-z0-9])_?([A-Z])", r"_\1", string)
    return string.lower()


# keys are valid specified MDF keys
# values are valid attributes of a bento_meta.Entity
mdf_to_meta = {
    "Handle": "handle",
    "Nodes": None,
    "Relationships": None,
    "PropDefinitions": None,
    "UniversalNodeProperties": None,
    "UniversalRelationshipProperites": None,
    "Terms": None,
    "Term": None,
    "Desc": "desc",
    "NanoID": "nanoid",
    "Src": "src",
    "Dst": "dst",
    "Mul": "multiplicity",
    "Value": "value",
    # don't translate type spec in init - process in process_prop
    "Type": "Type",
    "Enum": "Enum",
    "Key": "is_key",
    "Nul": "is_nullable",
    "Req": "is_required",
    "Deprecated": "is_deprecated",
    "Strict": "is_strict",
    "Origin": "origin_name",
    "Definition": "origin_definition",
    "Code": "origin_id",
    "Version": "origin_version",
}

# scalar types - shall match mdf_schema["defs"]["simpleType"]["enum"]
simple_types = [
    "number",
    "integer",
    "string",
    "datetime",
    "url",
    "boolean",
    "TBD",
]

# "declare"
process = {}


# hdl: handle, spec: MDF specifier object (from YAML)
# init: additional (bento-meta) attributes,
# entCls: bento-meta entity class (bareword)


def spec_to_entity(
    hdl: str | None,
    spec: dict,
    init: dict,
    ent_cls: type[Node | Edge | Property | Term | Tag],
) -> Entity:
    """Translate part of MDF YAML to bento-meta entity."""
    for k in spec:
        if k in mdf_to_meta and mdf_to_meta[k] is not None:
            init[mdf_to_meta[k]] = spec[k]
    if hdl and "handle" not in init:
        init["handle"] = hdl
    # init now contains (translated) spec keys and its original keys
    ent = ent_cls(init)
    info = process[ent_cls](init, ent)
    if spec.get("Tags"):
        for t in spec["Tags"]:
            ent.tags[t] = spec_to_entity(
                "",
                {},
                {"key": t, "value": spec["Tags"][t], "_commit": init.get("_commit")},
                Tag,
            )
    return ent


def process_node(spec: dict, node: Node) -> None:
    """Additional processing for Node entities."""
    pass

def process_reln(spec: dict, edge: Edge) -> None:
    """Additional processing for Relationship entities."""
    pass

def process_term(spec: dict, term: Term) -> None:
    """Additional processing for Term entities."""
    if not term.handle:
        term.handle = to_snake_case(term.value)
    if spec.get("definition"):
        term.definition = unquote(term, spec["definition"])


def process_tag(spec: dict, tag: Tag):
    """Additional processing for tag entities."""
    pass

def process_prop(spec: dict, prop: Property) -> None:
    """Additional processing for Property entities."""
    ty_spec = spec.get("Enum") or spec.get("Type")
    domain_spec = typespec_to_domain_spec(ty_spec)
    if (
        domain_spec["value_domain"] != "value_set"
        and domain_spec.get("item_domain") != "value_set"
    ):
        for attr in domain_spec:
            prop.__setattr__(attr, domain_spec[attr])
    else:  # is "value_set" or "list" with item_domain "value_set"
        if domain_spec["value_domain"] == "list":
            prop.value_domain = "list"
            prop.item_domain = "value_set"
        else:
            prop.value_domain = "value_set"
        if domain_spec.get("units"):
            prop.units = domain_spec["units"]
        if domain_spec.get("pattern"):
            prop.pattern = domain_spec["pattern"]
        if domain_spec.get("url"):
            prop.value_set = ValueSet(
                {"url": domain_spec["url"], "_commit": prop._commit},
            )
        elif domain_spec.get("path"):
            prop.value_set = ValueSet(
                {"path": domain_spec["path"], "_commit": prop._commit},
            )
        elif domain_spec.get("value_set"):
            # create 'dummy' ValueSet to hold Enum terms,
            # but merge them with any Terms section terms in mdf.py
            prop.value_set = ValueSet({"_commit": "dummy"})
            for tm_init in domain_spec["value_set"]:
                tm_init["origin_name"] = prop.model
                term = spec_to_entity(None, {"_commit": prop._commit}, tm_init, Term)
                prop.value_set.terms[term.handle] = term
        else:
            msg = f"Can't evaluate value_set spec {domain_spec}"
            raise RuntimeError(msg)
    # set default values for unprovided attrs
    prop_attrs_with_defaults = {
        "is_strict": True,
        "is_key": False,
        "is_nullable": False,
        "is_required": False,
    }
    for attr, default in prop_attrs_with_defaults.items():
        if getattr(prop, attr) is None:
            setattr(prop, attr, default)


process = {
    Node: process_node,
    Edge: process_reln,
    Property: process_prop,
    Term: process_term,
    Tag: process_tag,
}


def typespec_to_domain_spec(spec: str | dict | list) -> dict:
    # simple type
    if isinstance(spec, str):
        return {"value_domain": spec}
    if isinstance(spec, dict):
        # list type
        if spec.get("value_type") == "list":
            domain_spec = {"value_domain": "list"}
            list_spec = spec.get("item_type") or spec.get("Enum")
            item_domain_spec = typespec_to_domain_spec(list_spec)
            domain_spec["item_domain"] = item_domain_spec["value_domain"]
            for key in ("units", "value_set", "url", "path"):
                if key not in item_domain_spec:
                    continue
                domain_spec[key] = item_domain_spec[key]
            return domain_spec
        # regex type
        if spec.get("pattern"):
            # TODO: add regex flavor attribute to bento-meta Property entity?
            return {"value_domain": "regexp", "pattern": spec["pattern"]}
        # numberWithUnits type
        if spec.get("units"):
            domain_spec = {"value_domain": spec["value_type"]}
            if isinstance(spec["units"][0], dict) and spec["units"][0].get("pattern"):
                item_domain_spec = typespec_to_domain_spec(spec["units"][0])
                domain_spec["units"] = item_domain_spec["value_domain"]
                domain_spec["pattern"] = item_domain_spec["pattern"]
            else:
                domain_spec["units"] = ";".join(spec["units"])
            return domain_spec
    # enum type
    if isinstance(spec, list):
        # do not implement union type for now
        # assume a value_set, as a list of term handles or a single url/path
        # don't merge terms here, but in process_prop
        # list of term initializers returned
        #   term values == term handles
        if (
            len(spec) == 1
            and isinstance(spec[0], str)
        ):
            if re.match("^[a-z][a-z]*://",spec[0]):
                return {"value_domain": "value_set", "url": spec[0]}
            elif re.match("^/.*",spec[0]):
                return {"value_domain": "value_set", "path": spec[0]}
        vs = []
        for tm in spec:
            if isinstance(tm, bool):
                tm = "True" if tm else "False"  # stringify any booleans
            vs.append({"handle": tm, "value": tm})
        return {"value_domain": "value_set", "value_set": vs}
    # unknown - default domain
    return {"value_domain": Property.default("value_domain")}


def entity_to_spec(
        ent: Entity,
        spec: dict = None
) -> dict:
    if spec is None:
        spec = {}
    if ent.tags:
        tags = {}
        for t in ent.tags.values():
            entity_to_spec(t,tags)
        spec["Tags"] = tags
    if "concept" in ent.attspec and ent.concept and ent.concept.terms:
        spec["Term"] = [entity_to_spec(t) for t in ent.concept.terms.values()]
    if ent.desc:
        spec["Desc"] = ent.desc
    if isinstance(ent, Node):
        spec["Props"] = sorted([x.handle for x in ent.props.values()])
        if len(spec["Props"]) == 0:
            spec["Props"] = "null"
        pass
    elif isinstance(ent, Property):
        if ent.is_nullable is not None:
            spec["Nul"] = ent.is_nullable
        if ent.is_required is not None:
            spec["Req"] = ent.is_required
        if ent.is_strict is not None:
            spec["Strict"] = ent.is_strict
        if ent.is_key is not None:
            spec["Key"] = ent.is_key
        if ent.is_deprecated is not None:
            spec["Deprecated"] = ent.is_deprecated
        if ent.value_domain == 'value_set':
            spec["Enum"] = domain_spec_to_typespec(ent)
        else:
            spec["Type"] = domain_spec_to_typespec(ent)
        pass
    elif isinstance(ent, Edge):
        # note that edge mdf specs must be merged correctly in the caller
        if ent.is_required:
            spec["Req"] = True
        if ent.multiplicity:
            spec["Mul"] = ent.multiplicity
        spec["Src"] = ent.src.handle
        spec["Dst"] = ent.dst.handle
        spec["Props"] = sorted([x.handle for x in ent.props.values()])
        if len(spec["Props"]) == 0:
            spec["Props"] = "null"
        pass
    elif isinstance(ent, Term):
        spec["Value"] = ent.value
        spec["Origin"] = ent.origin_name
        if ent.origin_definition:
            spec["Definition"] = ent.origin_definition
        if ent.origin_version:
            spec["Version"] = ent.origin_version
        if ent.origin_id:
            spec["Code"] = ent.origin_id
        pass
    elif isinstance(ent, Tag):
        spec[ent.key] = ent.value
        pass
    else:
        raise RuntimeError(f'Cannot process Entity subtype: {ent}')
    return spec


def domain_spec_to_typespec(prop : Property) -> str | dict:
    ret = {}
    if prop.value_domain == 'value_set':
        if prop.value_set.url:
            ret = [prop.value_set.url]
        elif prop.value_set.path:
            ret = [prop.value_set.path]
        else:
            ret = [x for x in prop.terms]
    elif prop.value_domain == 'list':
        ret["value_type"] = 'list'
        if prop.item_domain == "value_set":
            if prop.value_set.url:
                ret["Enum"] = [prop.value_set.url]
            elif prop.value_set.path:
                ret["Enum"] = [prop.value_set.path]
            else:
                ret["Enum"] = [x for x in prop.terms]
        else:
            if prop.units:
                ret["item_type"] = {"value_type": prop.item_domain,
                                    "units": prop.units.split(";")}
            elif prop.pattern:
                ret["item_type"] = {"pattern": prop.pattern}
            else:
                ret["item_type"] = prop.item_domain
    elif prop.units:
        ret["value_type"] = prop.value_domain
        if prop.units == "regexp":
            ret["units"] = [{"pattern": prop.pattern}]
        else:
            ret["units"] = prop.units.split(";")
    elif prop.pattern:
        ret["pattern"] = prop.pattern
    else:
        ret = prop.value_domain
    return ret
